{%- if show_source and has_source and sourcename %}
  <h3>{{ _('This Page') }}</h3>
  <div>
    <a href="{{ pathto('_sources/' + sourcename, true)|e }}"
       rel="nofollow">{{ _('Show Source') }}</a>
  </div>
{%- endif %}
