// spc bootstrap settings
@import "bootstrap/bootstrap.less";
@import "bootstrap/responsive.less";

// google webfont
@import url(https://fonts.googleapis.com/css?family=Open+Sans);

//Typography
@sansFontFamily: 'Open Sans', sans-serif !important;
@baseFontSize: 13px;
@baseLineHeight: 19px;

//Colors
@blue: #12567D;

//Sprites
@iconSpritePath: '../../img/glyphicons-halflings.png';
@iconWhiteSpritePath: '../../img/glyphicons-halflings-white.png';
