body {
  background-color: #f9faf5;
}
.container {
  width: 80%;
}
.main {
  background-color: white;
  padding: 18px;
  -moz-box-shadow: 0 0 3px #888;
  -webkit-box-shadow: 0 0 3px #888;
  box-shadow: 0 0 3px #888;
}
.underline {
  border-bottom: 1.5px solid #eeeeee;
}
.header {
  margin-top: 15px;
  margin-bottom: 15px;
  margin-left: 0px;
  margin-right: 0px;
}
.spc-navbar {
  margin-top: 15px;
  margin-bottom: 5px;
  margin-left: 0px;
  margin-right: 0px;
}
.spc-navbar .nav-pills {
  margin-bottom: 0px;
  font-size: 12px;
}
.spc-navbar .nav-pills > li > a {
  padding-top: 2.5px;
  padding-bottom: 2.5px;
}
.underline {
  border-bottom: 1.5px solid #eeeeee;
}
.spc-page-title h1,
.spc-page-title h2,
.spc-page-title h3,
.spc-page-title h4 {
  font-weight: normal;
  border-bottom: 1.5px solid #eeeeee;
}
.tags .btn {
  border: none;
  font-size: 9.5px;
  font-weight: bold;
}
.spc-search-result-title h1,
.spc-search-result-title h2,
.spc-search-result-title h3,
.spc-search-result-title h4 {
  font-weight: normal;
}
.spc-snippet-header {
  margin-bottom: 5px;
}
.spc-snippet-info {
  padding-top: 10px;
}
.spc-snippet-info .dl-horizontal {
  margin: 5px;
}
.spc-snippet-info .dl-horizontal dt {
  font-weight: normal;
}
.spc-snippet-body {
  padding: 10px;
}
.spc-snippet-body .accordion-group {
  border: none;
}
.spc-snippet-body .accordion-heading {
  text-transform: uppercase;
  font-size: 14px;
  border-bottom: 1px solid #e5e5e5;
}
.spc-snippet-body .accordion-heading .accordion-toggle {
  padding-top: 10px;
  padding-bottom: 5px;
}
.spc-rightsidebar {
  color: #555555;
}
.spc-rightsidebar .navigation {
  padding: 2px 10px;
  font-size: 11.9px;
}
.spc-rightsidebar .navigation .nav-title {
  font-weight: bold;
  text-transform: uppercase;
}
.spc-rightsidebar .navigation li {
  margin: 5px;
}
.footer {
  padding: 5px;
  font-size: small;
}
