// LESS Utilities for spc
@import "bootstrap/variables.less";

.padding (@top: 0px, @bottom: 0px, @left: 0px, @right: 0px) {
	padding-top: @top;
	padding-bottom: @bottom;
	padding-left: @left;
	padding-right: @right;
}

.margin (@top: 0px, @bottom: 0px, @left: 0px, @right: 0px) {
	margin-top: @top;
	margin-bottom: @bottom;
	margin-left: @left;
	margin-right:@right;
}

.underline {
	border-bottom: 1.5px solid @hrBorder;
}
